uses DOS;

{ Global constants }
const BUF_SIZE=61440;   { read buffer size           }
      CODE_LEN=44;      { code length in bytes       }
      BackupExt='.CRT'; { extension for backup files }
      MaskExt='*.EXE';  { default mask extension     }
      { code as it is in buggy files }
      BUG_CODE:array[0..CODE_LEN-1] of byte=($FF, $DC, $36, $8B, $4F, $00, $E3,
                                             $13, $8E, $06, $00, $00, $33, $FF,
                                             $26, $8A, $1D, $A1, $00, $00, $33,
                                             $D2, $E8, $05, $00, $E2, $F6, $CA,
                                             $02, $00, $2D, $01, $00, $83, $DA,
                                             $00, $72, $05, $26, $3A, $1D, $74,
                                             $F3, $C3);
      { code as it should be }
      FIX_CODE:array[0..CODE_LEN-1] of byte=($FF, $DC, $36, $8B, $4F, $04, $E3,
                                             $13, $8E, $06, $44, $00, $33, $FF,
                                             $26, $8A, $1D, $A1, $00, $00, $51,
                                             $E8, $06, $00, $59, $E2, $F6, $CA,
                                             $02, $00, $48, $B9, $2A, $00, $E2,
                                             $FE, $74, $05, $26, $3A, $1D, $74,
                                             $F3, $C3);

var s, mask: string;                       { temp string; search mask      }
    p: DirStr; n: NameStr; e: ExtStr;      { temp strings                  }
    buf: array[0..BUF_SIZE-1] of byte;     { read buffer                   }
    i, code, proc, patch, found: integer;  { some stuff;; number of processed,
                                             found and patched files       }
    IOerrs, del: word;                     { delay factor, # of I/O errors }
    backup, test, ask, log, recur: boolean;{ backup files, only test, ask
                                             before patching, write log file,
                                             recursive search              }
    patchonly: boolean;                    { log patched files only        }
    logfile: text;                         { log file                      }

{ ReadKey - just like in CRT }
function ReadKey: char;
begin
  asm
    mov ah, 0
    int 16h
    mov @Result, al
  end
end;

{ makes string consisting of given substring }
function MakeStr(sub: string; count: integer): string;
var res: string;
    i: integer;

begin
  res:='';
  for i:=1 to count do
      res:=res+sub;
  MakeStr:=res;
end;

{ writes integer in hex }
function IntToHex(val: longint): string;
const HexChars: array[0..15] of char='0123456789ABCDEF';

var res: string;
    i: integer;

begin
  res:='h';
  for i:=1 to 8 do begin
      res:=HexChars[val and $F]+res;
      val:=val shr 4;
  end;
  IntToHex:=res;
end;

{ auto-detect required delay factor }
function DelDetect: word;
begin
  asm
    { initialize }
    mov ax, 40h       { set ES:DI to $40:$6C }
    mov es, ax
    mov di, 6Ch
    mov dx, 0         { set DX (tested delay factor) to 0 }
@test:
    add dx, 2         { add to delay factor 2 }
    mov ax, 0FFFFh    { set AX to maximum     }
    { wait to the end of this tick }
    mov bl, es:[di]   { save current timer value }
@start:               { wait until it changes }
    cmp bl, es:[di]
    je @start
    { emulate Delay()'s work }
    mov bl, es:[di]   { save current timer value }
@loop:
    dec ax            { decrease AX }
    mov cx, dx        { make pause for a while }
@pause: loop @pause
    je @end
    cmp bl, es:[di]   { until timer changes }
    je @loop
@end:
    cmp ax, 0         { if AX is decreased to ZERO }
    je @test          { try bigger delay factor    }
    inc dx            { else return current        }
    mov @Result, dx
  end;
end;

{ Backups file }
procedure BackupFile(oldname: string);
var oldf, newf: file;
    size: longint;
    newname: string;

begin
  { make new name - with extension BackupExt }
  size:=Pos('.', oldname);                             { find dot's position }
  if i=0 then newname:=oldname+BackupExt               { dot not found       }
     else newname:=copy(oldname, 1, size-1)+BackupExt; { dot found           }
  { write info }
  Write('Backuping (', oldname, ' -> ', newname, ')... ');
  { try to open files }
  Assign(oldf, oldname);
  Assign(newf, newname);
  {$I-}
  Reset(oldf, 1);
  if IOResult<>0 then begin
     { failed }
     WriteLn('Can''t open file "'+oldname+'"!');
     inc(IOerrs);
     exit;
  end;
  Rewrite(newf, 1);
  if IOResult<>0 then begin
     { failed }
     WriteLn('Can''t create file "'+newname+'"!');
     inc(IOerrs);
     exit;
  end;
  {$I+}
  { read file and write it to new destination }
  while not EOF(oldf) do begin
    if FileSize(oldf)-FilePos(oldf)>BUF_SIZE then size:=BUF_SIZE
       else size:=FileSize(oldf)-FilePos(oldf);
    BlockRead(oldf, buf, size);
    BlockWrite(newf, buf, size);
  end;
  { close all }
  Close(newf);
  Close(oldf);
  { write info }
  WriteLn('Done.');
end;

{ try to fix file }
procedure Fix(fname: string);
var f: file;
    i, size, fpos, ps: longint;
    BugPos, FixPos: integer;
    key: char;
    attr: word;

label _FIX_;

begin
  { try to open file }
  Assign(f, fname);
  { readonly file }
  GetFAttr(f, attr);
  if attr and ReadOnly>0 then SetFAttr(f, attr - ReadOnly);
  {$I-}
  Reset(f, 1);
  if IOResult<>0 then begin
     { failed }
     WriteLn('Can''t open file "', fname, '"!');
     if log and not patchonly then Write(logfile, 'Can not open file!');
     inc(IOerrs);
     exit;
  end;
  {$I+}
  { set positions in BUGGY and FIXED code to 0 }
  BugPos:=0;
  FixPos:=0;
  { search for it... }
  Write('Searching for buggy subroutine... ');
  while not EOF(f) do begin
    { get size to read }
    if FileSize(f)-FilePos(f)>BUF_SIZE then size:=BUF_SIZE else
       size:=FileSize(f)-FilePos(f);
    { read file to buffer }
    BlockRead(f, buf, size);
    ps:=FilePos(f);
    { search in buffer }
    i:=0;
    while i<size do begin
        { on some positions we have to remember original contents }
        if (BugPos=5) or (BugPos=10) or (BugPos=11) or
           (BugPos=18) or (BugPos=19) then FIX_CODE[BugPos]:=buf[i];
        if (FixPos=5) or (FixPos=10) or (FixPos=11) or
           (FixPos=18) or (FixPos=19) then FIX_CODE[FixPos]:=buf[i];
        { check for original code }
        if (buf[i]=BUG_CODE[BugPos]) or (BugPos=5) or (BugPos=10)
           or (BugPos=11) or (BugPos=18) or (BugPos=19) then inc(BugPos) else
              if BugPos>0 then begin      { we should unroll }
                 if i<BugPos-1 then begin { out of buffer }
                    { change position in file }
                    Seek(f, FilePos(f)-size+i-BugPos+1);
                    { change position in fixed code }
                    FixPos:=FixPos-BugPos+1;
                    if FixPos<0 then FixPos:=0;
                    { set position in buggy code to 0 }
                    BugPos:=0;
                    { re-read }
                    break;
                 end;
                 { in buffer }
                 { change position in fixed code }
                 FixPos:=FixPos-BugPos+1;
                 if FixPos<0 then FixPos:=0;
                 { change position in buffer }
                 i:=i-BugPos+1;
                 { set position in buggy code to 0 }
                 BugPos:=0;
                 continue;
              end;
        { original code found }
        if BugPos=CODE_LEN then begin
           { write info }
           WriteLn('Found!');
           { remember position of code in file and go to fix it }
           fpos:=FilePos(f)-size+i-CODE_LEN+1;
           goto _FIX_;
        end;
        { check for patched code }
        if (buf[i]=FIX_CODE[FixPos]) or (FixPos=5) or (FixPos=10)
           or (FixPos=11) or (FixPos=18) or (FixPos=19) or (FixPos=32)
           or (FixPos=33) then inc(FixPos) else
              if FixPos>0 then begin       { we should unroll }
                 if i<FixPos-1 then begin  { out of buffer }
                    { change position in file }
                    Seek(f, FilePos(f)-size+i-FixPos+1);
                    { change position in buggy code }
                    BugPos:=BugPos-FixPos;
                    if BugPos<0 then BugPos:=0;
                    { set position in fixed code to 0 }
                    FixPos:=0;
                    { re-read }
                    break;
                 end;
                 { in buffer }
                 { change position in buggy code }
                 BugPos:=BugPos-FixPos;
                 if BugPos<0 then BugPos:=0;
                 { change position in buffer }
                 i:=i-FixPos+1;
                 { set position in fixed code to 0 }
                 FixPos:=0;
                 continue;
              end;
        { patched code found }
        if FixPos=CODE_LEN then begin
           { write info }
           WriteLn('Fixed code found!');
           { remember position of code in file and go to fix it }
           fpos:=FilePos(f)-size+i-CODE_LEN+1;
           goto _FIX_;
        end;
        inc(i);
    end;
  end;
  { nothing was found }
  WriteLn('Not found!');
  if log and not patchonly then Write(logfile, '    Not found       No        No');
  if attr and ReadOnly>0 then SetFAttr(f, attr); { restore readonly attrib }
  Close(f);
  exit;
  { fix code }
_FIX_:
  inc(found);
  Close(f);                              { close file                     }
  { logging }
  if log and patchonly then Write(logfile, FExpand(fname), MakeStr(' ', 46-Length(FExpand(fname))));
  if log then Write(logfile, IntToHex(fpos):13);
  if test then begin                     { test mode                      }
     if log then Write(logfile, '       No        No');
     if log and patchonly then WriteLn(logfile);
     { restore readonly attribute }
     if attr and ReadOnly>0 then SetFAttr(f, attr);
     exit;
  end;
  if ask then begin                      { ask for patching               }
     repeat                              { force user to press <Y> or <N> }
        if backup then                   { prompt                         }
           Write('Patch this file (Y/N/Q)? ')
        else
           Write('Patch this file (Y/N/B/Q)? ');
        key:=UpCase(ReadKey);
        WriteLn(key);
        { analys pressed key }
        if (key='B') and not backup then begin
           BackUpFile(fname);
           if log then Write(logfile, '      Yes       Yes');
           break;
        end;
        if key='N' then begin      { exit if we shouldn't patch this file }
           if log then Write(logfile, '       No        No');
           { restore readonly attribute }
           if attr and ReadOnly>0 then SetFAttr(f, attr);
           exit;
        end;
        if key='Q' then begin      { exit from program              }
           { restore readonly attribute }
           if attr and ReadOnly>0 then SetFAttr(f, attr);
           halt(0);
        end;
     until (key='Y');
  end;
  if backup then begin             { if we need to backup file            }
     BackUpFile(fname);
     if log then Write(logfile, '      Yes       Yes');
  end else if log then Write(logfile, '      Yes        No');
  { try to reopen file }
  Assign(f, fname);
  {$I-}
  Reset(f, 1);
  if IOResult<>0 then begin              { failed                         }
     WriteLn('Can''t open file "', fname, '"!');
     if log and not patchonly then Write(logfile, 'Can not open file!');
     inc(IOerrs);
     exit;
  end;
  {$I+}
  Seek(f, fpos);                      { go to code's position }
  { write info }
  if BugPos=CODE_LEN then Write('Fixing code... ')
     else Write('Refixing code... ');
  BlockWrite(f, FIX_CODE, CODE_LEN);  { fix code              }
  WriteLn('Fixed.');                  { write info            }
  if attr and ReadOnly>0 then SetFAttr(f, attr); { restore readonly attrib }
  Close(f);                           { close file            }
  inc(patch);                         { one more file fixed   }
  { logging }
  if log and patchonly then WriteLn(logfile);
end;

procedure Search;
var srch: SearchRec;    { for searching     }

begin
  { try to find first matching file }
  FindFirst(mask, AnyFile-VolumeID-Directory, srch);
  while DosError=0 do begin  { while we can find anything }
      WriteLn('Processing file "', FExpand(srch.name), '".');  { write info }
      { logging }
      if log and not patchonly then Write(logfile, FExpand(srch.name), MakeStr(' ', 46-Length(FExpand(srch.name))));
      Fix(srch.name);      { try to fix file          }
      WriteLn('Done.');    { write info               }
      WriteLn;
      { logging }
      if log and not patchonly then WriteLn(logfile);
      inc(proc);           { one more file is checked }
      FindNext(srch);      { try to find next         }
  end;
  if recur then begin      { if resursive search - find all dirictories }
     FindFirst('*.*', AnyFile, srch);
     while DosError=0 do begin  { while we can find anything }
        { if this is real directory }
        if (srch.attr and Directory>0) and (srch.name<>'.') and (srch.name<>'..') then begin
            {$I-}
            ChDir(srch.name);   { change directory   }
            if IOResult<>0 then begin
               WriteLn('Can not change directory to ''', srch.name, '''!');
               inc(IOerrs);
               ReadKey;
               continue;
            end;
            {$I+}
            Search;             { search in current  }
            ChDir('..');        { restore directory  }
        end;
        FindNext(srch);         { try to find next           }
     end;
  end;
end;

{ Teachs user }
procedure GetHelp;
begin
  WriteLn('USAGE: CRTFix [options] <file_mask> [options]');
  WriteLn('  where: file_mask is mask of files to be patched and options are:');
  WriteLn('         /a - ask before patching;');
  WriteLn('         /b - backup files;');
  WriteLn('         /d## - delay factor, ##: 0 - autodetect (default), else - set');
  WriteLn('                delay factor to ## (range 1 - 65535). Smaller delay');
  WriteLn('                factor gives more precision, greater - allows to run');
  WriteLn('                programm on faster CPUs;');
  WriteLn('         /l - write log file;');
  WriteLn('         /p - write to log only files, containing CRT code;');
  WriteLn('         /r - recursive search;');
  WriteLn('         /t - test mode: only search for code, but don''t patch;');
  WriteLn('         /h or /? - this help.');
  WriteLn;
  halt(0);
end;

begin
  { Write header }
  WriteLn;
  WriteLn('CRTFix version 1.10');
  WriteLn('Copyright (C) 2000 by Eugene Toder.');
  WriteLn('Fixes error that cases run-time error 200 in programms that uses CRT unit.');
  WriteLn('Can be applied to both EXEs and TPUs, TPLs.');
  WriteLn;
  { set default values }
  del:=0;               { auto-detect delay factor   }
  backup:=false;        { don't backup files         }
  test:=false;          { don't just test            }
  ask:=false;           { don't ask before patching  }
  log:=false;           { don't write log            }
  patchonly:=false;     { log patched files only     }
  recur:=false;         { don't use recursive search }
  mask:='';             { empty search mask          }
  { read command-line parameters }
  for i:=1 to ParamCount do begin
      s:=ParamStr(i);             { read parameter }
      { if parameter starts with / or - it's a switch, else - a search mask }
      if (s[1]<>'/') and (s[1]<>'-') then begin
         { file mask }
         if mask=''then mask:=s { save mask }
         else begin
            { duplicate mask }
            WriteLn('Duplicate file mask!');
            WriteLn;
            GetHelp;
         end;
      end else begin
         if (Length(s)>2) and (s[2]<>'d') then begin { wrong long switch }
            WriteLn('Invalid parameter: "'+s+'"!');
            WriteLn;
            GetHelp;
         end;
         case s[2] of                             { analys switch           }
           'a': ask:=true;                        { /a - ask                }
           'b': backup:=true;                     { /b - backup files       }
           'd': begin                             { /d - delay factor       }
                  Val(copy(s, 3, 100), del, code);
                  if code<>0 then begin           { error converting to int }
                     WriteLn('Invalid delay factor: "'+copy(s, 3, 100)+'"!');
                     WriteLn;
                     GetHelp;
                  end;
                end;
           'l': log:=true;                        { /l - write log file     }
           'p': patchonly:=true;                  { /p - log patched only   }
           'r': recur:=true;                      { /r - recursive search   }
           'h', '?': GetHelp;                     { /h, /? show help        }
           't': test:=true;                       { /t - test mode          }
         else
           { unknown switch }
           WriteLn('Invalid parameter: "'+s+'"!');
           WriteLn;
           GetHelp;
         end;
      end;
  end;
  if mask='' then begin { mask wasn't specified }
     WriteLn('Not enough actual parameters!');
     WriteLn;
     GetHelp;
  end else
      { if mask isn't full (dont have '.') - extend it with '*.exe' }
      if Pos('.', mask)=0 then mask:=mask+MaskExt; { not found }
  if (del=0) and not test then begin { auto-detect delay factor }
     Write('Detecting delay factor... ');
     del:=DelDetect;
     WriteLn('Done.');
  end;
  { write info }
  if test then WriteLn('Test mode.') else
     WriteLn('Delay factor is ', del, '.');
  WriteLn;
  { place delay factor in code }
  FIX_CODE[32]:=Lo(del);
  FIX_CODE[33]:=Hi(del);
  { anti-self-patching trick }
  BUG_CODE[0]:=$8B;
  FIX_CODE[0]:=$8B;
  { nothing found or patched yet }
  proc:=0;
  patch:=0;
  found:=0;
  IOerrs:=0;    { not I/O errors yet }
  { logging }
  if log then begin
     Assign(logfile, 'crtfix.log');
     {$I-}
     Rewrite(logfile);
     if IOResult<>0 then begin
        WriteLn('Can not create log file!');
        inc(IOerrs);
        exit;
     end;
     {$I+}
     WriteLn(logfile, 'FILE NAME                                     CODE POSITION  PATCHED  BACKUPED');
     WriteLn(logfile, '------------------------------------------------------------------------------');
  end;
  { search for files }
  FSplit(mask, p, n, e);
  mask:=n+e;
  GetDir(0, s);
  {$I-}
  ChDir(p);
  if IOResult<>0 then begin
     WriteLn('Can not change directory to ''', p, '''!');
     inc(IOerrs);
     exit;
  end;
  {$I+}
  Search;
  ChDir(s);
  { write statictis }
  WriteLn(IOerrs, ' I/O error(s).');
  if proc=0 then WriteLn('No files found!')
     else WriteLn(proc, ' file(s) processed, code found in ', found, ' file(s), ', patch, ' file(s) patched.');
  { logging }
  if log then begin
     WriteLn(logfile, IOerrs, ' I/O error(s).');
     if proc=0 then WriteLn(logfile, 'No files found!')
        else WriteLn(logfile, proc, ' file(s) processed, code found in ', found, ' file(s), ', patch, ' file(s) patched.');
  end;
  { close log }
  if log then Close(logfile);
  WriteLn;
end.